'' Display API for use with either the VgaDriver or TvDriver.

'' 2007-12-21 - Initial implementation
'' 2007-12-23 - Added windowEditor
'' 2007-12-30 - Changed to DAT for globals

OBJ
   def : "Definitions"           ' Common definitions

CON
   romFont   = $8000             ' Start of ROM font table
   blankTile = romFont + " "<<6  ' Tile value for blank
   tabCols   = 8                 ' Columns for each tab

DAT
screen  long  0                  ' Address of tile array
colors  long  0                  ' Address of color table
tiles   long  0                  ' Number of display tiles
cols    long  0                  ' Number of columns on display
rows    long  0                  ' Number of rows on display
mode    long  0                  ' Tile mode (0-16x16, 1-16x32)
colNo   long  0                  ' Current column number
rowNo   long  0                  ' Current row number
color   long  0                  ' Current color (0-63)
arg     long  0                  ' For multi-byte control sequences

PUB start
   screen := long[def#dspArray]  ' Get address of tile array
   colors := long[def#dspColor]  ' Get address of color table
   tiles := long[def#dspTiles]   ' Get number of display tiles
   cols := long[def#dspCols]     ' Get number of horizontal characters
   rows := long[def#dspRows]     ' Get number of vertical characters
   mode := long[def#dspMode]     ' Get tile mode (0-16x16,1-16x32)
   colNo := 0
   rowNo := 0
   color := 0
   arg := 0

PUB str(stringptr)

'' Print a zero-terminated string

  repeat strsize(stringptr)
    out(byte[stringptr++])

PUB dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    out("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10

PUB hex(value, digits)

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")

PUB out(c) | t, i                ' Display the character
   if c == def#Alt00
      c := 0                     ' Substitute $00 for Alt00
   case arg
      0:
         case c
            def#ClrScr:          ' Clear the screen to blank
               fillBlanks(0,0,tiles)
               colNo := rowNo := 0
            def#Home:            ' Set screen position to home
               colNo := rowNo := 0
            def#CursLt:          ' Move cursor to the left
               colNo := colNo - 1 #> 0
            def#CursRt:          ' Move cursor to the right
               colNo := colNo + 1 <# cols - 1
            def#CursUp:          ' Move cursor upwards
               rowNo := rowNo - 1 #> 0
            def#CursDn:          ' Move cursor down
               rowNo := rowNo + 1 <# rows - 1
            def#Bsp:             ' Backspace
               if colNo > 0
                  setChar(--colNo,rowNo," ",color)
            def#Tab:
               repeat
                  setChar(colNo++,rowNo," ",color)
                  if colNo == cols
                     colNo := cols - 1
                     quit
               while colNo // tabCols <> 0
            def#Lf:
               newline
            def#Insert:          ' Insert a blank at cursor
               if colNo <> cols - 1
                  t := screen + ((rowNo*cols)<<1)
                  repeat i from cols-1 to colNo+1
                     word[t][i] := word[t][i-1]
               setChar(colNo,rowNo," ",color)
            def#Cr:              ' Carriage return
               colNo := 0
            def#Delete:          ' Delete the charater at x,y
               if colNo <> cols - 1
                  t := screen + ((rowNo*cols)<<1)
                  repeat i from colNo to cols-2
                     word[t][i] := word[t][i+1]
               setChar(cols-1,rowNo," ",color)
            def#Color, def#ClrEol, def#MoveX, def#MoveY, def#MoveXY:
               arg := c
            other:               ' Displayable character
               setChar(colNo,rowNo,c,color)
               if ++colNo == cols ' Wraparound to next line
                  colNo := 0
                  newline
      def#Color:                 ' Save color number
         color := c <# 63
         arg := 0
      def#ClrEol:                ' Clear specified # columns up to EOL
         fillBlanks(colNo,rowNo,c <# cols-colNo)
         arg := 0
      def#MoveX:                 ' Move to specified column
         colNo := c <# cols-1
         arg := 0
      def#MoveY:                 ' Move to specified row
         rowNo := c <# rows-1
         arg := 0
      def#MoveXY:                ' Move to specified column,row
         colNo := c <# cols-1
         arg := def#MoveY

PRI newline                      ' Move to the next new line
   if ++rowNo == rows            ' Scroll up if at last row
      wordmove(screen,screen+(cols<<1),tiles-cols)
      rowNo := rows - 1
      fillBlanks(0,rowNo,cols)

PRI fillBlanks(x,y,size)         ' Fill screen with blanks from x,y
   wordfill(screen+((y*cols+x)<<1),blankTile+(color<<1),size)

PUB getChar(x,y)                 ' Return the character at x,y
   if (result := word[screen][y*cols+x]) > $7FFF
      return ((result-romFont) >> 6) & $FE | (result-romFont) & 1

PUB setChar(x,y,c,clr)           ' Set the character at x,y
   c := (c & $FE) << 6 | clr << 1 | c & 1
   word[screen][y*cols+x] := c + romFont

PUB getColor(x,y)                ' Return the color at x,y
   if (result := word[screen][y*cols+x]) > $7FFF
      return ((result-romFont) >> 1) & $3F

PUB col                          ' Return the current column
   return colNo

PUB row                          ' Return the current row
   return rowNo

PUB windowEditor(winX, winY, winC, winR, state) | c, s, clr

' This method provides an editable window on the display text screen.
' It uses the standard PS/2 keyboard control keys including backspace,
' cursor positioning, delete and insert (for an insert mode).
' The edited text must be retrieved from the window by the caller.
' Unknown or invalid control keys terminate editing and return the keycode.
' The state of the window and the screen is such that editing can be
' resumed after the terminating keycode is processed by the caller.
' The return value contains: insert << 24 | y << 16 | x << 8 | keycode
' and is normally used on a subsequent call to maintain the state.
' A value of zero is used to initialize the editor to the upper left
' corner of the window.  No checking is done for valid window bounds.
' The keyboard buffer and pointers are accessed directly so the rest of
' the keyboard driver is not needed if this is used for keyboard input.

   if state == 0
      repeat c from 0 to winR-1  ' Clear the window to blanks
         fillBlanks(winX,winY+c,winC)
   colNo := ((state>>8)&$FF)+winX ' Reposition the cursor
   rowNo := ((state>>16)&$FF)+winY
   state ~>= 24
   repeat
      s := getChar(colNo,rowNo)
      clr := getColor(colNo,rowNo)
      setChar(colNo,rowNo,$5F,clr) ' Display underline
      repeat while long[def#keyTail] == long[def#keyHead]
         ?long[def#randomSeed]   ' Wait for a key press
      c := word[def#keyKeys][long[def#keyTail]]
      long[def#keyTail] := (long[def#keyTail] + 1) & def#keyMask
      setChar(colNo,rowNo,s,clr) ' Restore character      
      case c
         $09:                    ' Tab
            repeat
               if state          ' Insert spaces to next tab stop
                  out(def#Insert) ' and always insert at least one
               out(" ")
            while (colNo - winX) // tabCols
         $0D:                    ' Return
            out(def#ClrEol)      
            out(winC)            ' Clear to end of line and go
            out(def#Cr)          '  to beginning of next line
            if rowNo - winY == winR - 1 ' Exit if already at bottom of window
               return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c
            out(def#Lf)
         $20..$7E:               ' Printable characters
            if state
               out(def#Insert)
            out(c)
         $C0:                    ' Left arrow
            out(def#CursLt)
         $C1:                    ' Right arrow 
            out(def#CursRt)                            
         $C2:                    ' Up arrow
            if rowNo == winY     ' Exit if already at top of window
               return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c
            if winR == 1         ' Special case for a single line window
               return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c
            out(def#CursUp)
         $C3:                    ' Down arrow
            if rowNo - winY == winR - 1 ' Exit if already at bottom of window
               return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c
            if winR == 1         ' Special case for a single line window
               return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c
            out(def#CursDn)
         $C4:                    ' Home
            if winR == 1         ' Special case for a single line window
               return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c
            out(def#Home)
         $C5:                    ' End
            if winR == 1         ' Special case for a single line window
               return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c
            out(def#MoveXY)
            out(0)
            out(winY+winR-1)
         $C8:                    ' Backspace
            if colNo > winX      ' Ignore if already at left side of window
               out(def#Bsp)
         $C9, $7E:               ' Delete
            out(def#Delete)
         $CA:                    ' Toggle Insert mode
            state := not state
         other:                  ' All others, exit
            return state<<24 | (rowNo-winY)<<16 | (colNo-winX)<<8 | c

PUB getLine(ptr,winX,winY,winC)  ' Copy the display line to the string
   repeat while winC > 0         '  and suppress trailing spaces
      if getChar(winX+winC-1,winY) <> " "
         quit
      winC--
   repeat winC
      byte[ptr++] := getChar(winX++,winY)
   byte[ptr]~                    ' Zero terminator

PUB scrollUp(winX,winY,winC,winR) | i, t ' Scroll up window
   if winR > 1
      repeat i from winY to winY+winR-2
         t := screen+((i*cols+winX)<<1)
         wordmove(t,t+(cols<<1),cols)
   fillBlanks(winX,winY+winR-1,winC)

PUB scrollDown(winX,winY,winC,winR) | i, t ' Scroll down window
   if winR > 1
      repeat i from winY+winR-2 to winY
         t := screen+((i*cols+winX)<<1)
         wordmove(t+(cols<<1),t,cols)
   fillBlanks(winX,winY,winC)
   