' This is the compiler for Ouroboros, a dialect of Basic based on PBasic.
' It has been modified and extended to support the Large Memory Model
' Propeller processor.

' Scalar 32-bit variables can be allocated in the local memory of the cog
' and scalar and array 8-bit, 16-bit, and 32-bit variables can be allocated
' in the global hub memory.

' 2007-12-17 - Corrected findSymbol
' 2007-12-18 - Corrected various routines for fixups
'              Added support for file includes
'              Added support for testing whether identifiers are defined
'              Added code generators for Boolean and relational operators

CON
   _xinfreq  = 5_000_000
   _clkmode  = xtal1 + pll16x
   _free     = 768+64+64

   version   = 1
   release   = 001
   testlevel = 2

   maxScan   = 255               ' Maximum scan buffer
   maxDict   = 512               ' Maximum number of identifiers and strings
   maxUnary  = 16                ' Maximum sequential unary operators
   initClk   = 12_000_000        ' Initial clock frequency (approx. RCFAST)
   
#0, tokPlus2, tokMinus2, tokUp2, tokBar2, tokStar2, tokSlash2, tokMaxOp
 tokMinOp, tokSAR, tokSet, tokSHL, tokSHR, tokROR, tokROL, tokDecode
 tokEncode, tokBitRev, tokGEQ, tokLEQ, tokNEQ, tokDotDot, tokAndNot, tokDefine 
 tokPlus, tokMinus, tokBitXor, tokBar, tokStar, tokSlash, tokClear, tokRandom
 tokAddr, tokBitNot, tokGTR, tokLSS, tokBitAnd, tokColon, tokLtPar, tokRtPar
 tokLtBrkt, tokRtBrkt, tokComma, tokSemi, tokPound, tokEQL, tokDot, tokBackSlash

 tokCON, tokVAR, tokBYTE, tokWORD, tokLONG, tokOR, tokAND, tokNOT, tokBRANCH
 tokIF, tokTHEN, tokELSE, tokENDIF, tokGOTO, tokGOSUB, tokRETURN, tokFOR
 tokTO, tokSTEP, tokNEXT, tokREAD, tokWRITE, tokLET, tokLOOKUP, tokLOOKDOWN
 tokINPUT, tokOUTPUT, tokREVERSE, tokLOW, tokHIGH, tokTOGGLE, tokPAUSE
 tokWAITCNT, tokWAITPNE, tokWAITPEQ, tokWAITVID, tokEND, tokDEBUG, tokCNT
 tokINA, tokOUTA, tokDIRA, tokCTRA, tokCTRB, tokFRQA, tokFRQB, tokPHSA
 tokPHSB, tokVCFG, tokVSCL, tokREBOOT, tokLOCKSET, tokLOCKCLR, tokLOCKNEW
 tokLOCKRET, tokCOGNEW, tokCOGINIT, tokCOGID, tokCOGSTOP, tokCLKSET, tokLOCAL
 tokINCLUDE

 tokChar, tokNumb, tokIdent, #221, tokLastId
 tokStr, #238, tokExtId, #254, tokEOL, tokEOF
   
   locals    = $080              ' Start of local variables in cog
   lastLocal = $1DF              ' End of local variables in cog
   temps     = 16                ' Number of temporary local values
   maxIncl   = 16                ' Number of file include references
   
   symStr    = 0                 ' String constant (stored in hub memory)
   symByte   = 1                 ' Byte variable (stored in hub memory)
   symWord   = 2                 ' Word variable (stored in hub memory)
   symLong   = 3                 ' Long variable (stored in hub memory)
   symLocl   = 4                 ' Long variable (stored in cog memory)
   symCons   = 5                 ' Named constant (symLen/symChn is value)
   symLabel  = 6                 ' Program label
   symAny    = 7                 ' Any identifier (not a string)
   symVal    = 8                 ' Used in stack for constant values
   symConId  = 9                 ' Used in stack for named constant
   
   always    = %1111             ' Condition codes
   never     = %0000
   ifC       = %1100
   ifNC      = %0011
   ifZ       = %1010
   ifNZ      = %0101
   WrZ       = %1000             ' Effects flags
   WrC       = %0100
   WrR       = %0010
   Imm       = %0001
   PC        = 0                 ' LMM program counter
   lmmCall   = 3                 ' LMM subroutine call
   lmmRetn   = 4                 ' LMM subroutine return
   lmmCache  = 5                 ' LMM cache load & execute
   lmmLoad   = 6                 ' LMM load immediate long
   lmmMult   = 7                 ' LMM unsigned 32 bit multiply
   lmmDiv    = 8                 ' LMM unsigned 32 bit divide
   lmmOnes   = 9                 ' LMM constant for all 1 bits
   lmmTemp1  = 10                ' LMM temporary location 1
   lmmTemp2  = 11                ' LMM temporary location 2
 
OBJ lib : "Meta2Lib"

VAR
   long compiling                ' True if compiling statements
   byte stackType[temps]         ' Type of stack entry
   long stackVal[temps]          ' Stack values
   word localVars                ' Next available cog variable address
   word stackPtr                 ' Next available cog temporary address
   word clkmodePtr               ' Dictionary entry for _CLKMODE
   long clkmodeSet
   word clkfreqPtr               ' Dictionary entry for _CLKFREQ
   long clkfreqSet
   word dictPtr                  ' Next available dictionary entry
   word dictSym[maxDict]         ' Packed symbol number and symbol type
   word dictLen[maxDict]         ' Storage (in bytes) used by the symbol
   word dictChn[maxDict]         ' Start of linked list of references
   byte scanPtr                  ' Next available scanBuf location
   byte scanBuf[maxScan+1]       ' Buffer for scanned tokens
   word includes[maxIncl]        ' Token numbers of include strings

PUB Main_Meta2
   if lib.start(-1,true)
      if lib.checkAbort(\Ouroboros)
         lib.stop(-1)
      lib.exit

PRI loadAddr(n) | s, t, fix      ' Generate load for address in the stack
   s := locals - 1 - (stackPtr - n)
   t := stackVal[stackPtr-n]
   case stackType[stackPtr-n]
      symByte, symWord, symLong:
         if t =< maxDict
            if t == maxDict
               abort string("loadAddr - Already an address")
            fix := dictChn[t]
            dictChn[t] := lib.getOutputAddr + 4
            loadImmed(s,fix)
         else
            outputInst($28,WrR+Imm,always,s,t-maxDict-1) ' MOV #
      other:
         abort string("loadAddr - Not a hub variable")
   stackVal[stackPtr-n] := maxDict

PRI storeVal(n) | t, v, s, z, fix ' Store the top of the stack
   z := loadVals(1)
   t := stackType[stackPtr-n]
   v := stackVal[stackPtr-n]
   s := locals - 1 - (stackPtr - n)
   case t
      symByte, symWord, symLong:
         if v =< maxDict
            if v < maxDict
               fix := dictChn[v]
               dictChn[v] := lib.getOutputAddr + 4
               loadImmed(s,fix)
            case t
               symByte:
                  outputInst($00,0,always,z,s) ' WRBYTE
               symWord:
                  outputInst($01,0,always,z,s) ' WRWORD
               symLong:
                  outputInst($02,0,always,z,s) ' WRLONG
         else
            v -= maxDict + 1
            case t
               symByte:
                  outputInst($00,Imm,always,z,v) ' WRBYTE #
               symWord:
                  outputInst($01,Imm,always,z,v) ' WRWORD #
               symLong:
                  outputInst($02,Imm,always,z,v) ' WRLONG #
      symLocl:
         if v < maxDict
            outputInst($28,WrR,always,dictChn[v],z) ' MOV
         else
            if v-maxDict-1 <> s
               outputInst($28,WrR,always,v-maxDict-1,z) ' MOV
      other:
         abort string("storeVal - Invalid type")

PRI loadVals(n) | t, v, s, fix   ' Generate loads for values in the stack
   result := locals - 1 - (stackPtr - n)
   repeat while n > 0
      t := stackType[stackPtr-n]
      v := stackVal[stackPtr-n]
      s := locals - 1 - (stackPtr-n)
      case t
         symByte, symWord, symLong:
            if v =< maxDict
               if v < maxDict
                  fix := dictChn[v]
                  dictChn[v] := lib.getOutputAddr + 4
                  loadImmed(s,fix)
               case t
                  symByte:
                     outputInst($00,WrR,always,s,s) ' RDBYTE
                  symWord:
                     outputInst($01,WrR,always,s,s) ' RDWORD
                  symLong:
                     outputInst($02,WrR,always,s,s) ' RDLONG
            else
               v -= maxDict + 1
               case t
                  symByte:
                     outputInst($00,WrR+Imm,always,s,v) ' RDBYTE #
                  symWord:
                     outputInst($01,WrR+Imm,always,s,v) ' RDWORD #
                  symLong:
                     outputInst($02,WrR+Imm,always,s,v) ' RDLONG #
         symLocl:
            if v < maxDict
               outputInst($28,WrR,always,s,dictChn[v]) ' MOV
            else
               if v-maxDict-1 <> s
                  outputInst($28,WrR,always,s,v-maxDict-1) ' MOV
         symVal:
            case v
               0..511:
                  outputInst($28,WrR+Imm,always,s,v) ' MOV #
               -511..-1:
                  outputInst($29,WrR+Imm,always,s,-v) ' NEG #
               other:
                  loadImmed(s,v)
         other:
            abort string("loadVals - Invalid type")
      stackType[stackPtr-n] := symLocl
      stackVal[stackPtr-n--] := s + maxDict + 1

PRI pushVal(t,v)                 ' Push a value into the stack
   if stackPtr == temps
      abort string("pushVal - Stack overflow")
   if t == symCons               ' Push value of named constant
      t := symVal
      v := dictChn[v] << 16 | dictLen[v]
   stackType[stackPtr] := t
   stackVal[stackPtr++] := v

PRI checkCons(n)                 ' Returns true if specified stack
   repeat while n > 0            '  levels are all constant values
      if stackType[stackPtr - n--] <> symVal
         return false
   return true

PRI popVal
   if stackPtr-- == 0
      abort string("popVal - Stack underflow")

PRI mustBeVal
   if stackPtr-- == 0
      abort string("mustBeVal - Stack underflow")
   if stackType[stackPtr] <> symVal
      abort string("mustBeVal - Not a constant")
   return stackVal[stackPtr]

PRI getLabel | t, s
   case t := scanBuf[scanPtr++]
      tokIdent..tokLastId, tokExtId..tokExtId+15:
         if t < tokExtId
            t -= tokIdent
         else
            t -= tokExtId
            t := t << 8 + scanBuf[scanPtr++]
            t += tokLastId - tokIdent + 1
         case findSymbol(symAny,t) ' Get type of Id
            symAny, symLabel:
               return findSymbol(symLabel,t)
            other:
               abort string("getLabel - Not a label")
      other:
         abort string("getLabel - Not a label")

PRI variable | t, s
   case t := scanBuf[scanPtr++]
      tokIdent..tokLastId, tokExtId..tokExtId+15:
         if t < tokExtId
            t -= tokIdent
         else
            t -= tokExtId
            t := t << 8 + scanBuf[scanPtr++]
            t += tokLastId - tokIdent + 1
         case s := findSymbol(symAny,t) ' Get type of Id
            symAny:
               abort string("variable - Undefined variable")
            symByte, symWord, symLong, symLocl:
               pushVal(s,findSymbol(s,t))
            symCons:
               abort string("variable - Can't change constant")
         return true
      other:
         scanPtr--
         return false
         
PRI termLvl0 | t, s
   case t := scanBuf[scanPtr++]
      tokDefine:
         case scanBuf[scanPtr++]
            tokIdent..tokLastId, tokExtId..tokExtId+15:
               if t < tokExtId
                  t -= tokIdent
               else
                  t -= tokExtId
                  t := t << 8 + scanBuf[scanPtr++]
                  t += tokLastId - tokIdent + 1
               t := findSymbol(symAny,t) <> symAny
               pushVal(symVal,t) ' true if defined, false otherwise
            other:
               abort string("termLvl0 - ?? then not identifier")
      tokChar:
         s := strsize(@scanBuf+scanPtr)
         t := 0
         bytemove(@t,@scanBuf+scanPtr,s)
         scanPtr += s + 1
         pushVal(symVal,t)
      tokNumb:
         bytemove(@t,@scanBuf+scanPtr,4)
         scanPtr += 4
         pushVal(symVal,t)
      tokIdent..tokLastId, tokExtId..tokExtId+15:
         if t < tokExtId
            t -= tokIdent
         else
            t -= tokExtId
            t := t << 8 + scanBuf[scanPtr++]
            t += tokLastId - tokIdent + 1
         s := findSymbol(symAny,t) ' Get type of Id
         if s == symAny
            abort string("termLvl0 - Undefined variable")
         t := findSymbol(s,t)    ' Get dictionary entry for name
         pushVal(s,t)
      tokStr..tokStr+15:
         t := (t-tokStr) << 8 + scanBuf[scanPtr++]
         t := findSymbol(symStr,t)
         if dictLen[t] == 0      ' Get length if not set yet
            dictLen[t] := lib.outputSymbol(t,false) + 1
         pushVal(symByte,t)
      other:
         abort string("termLvl0 - Syntax error")

PRI termLvl1
   termLvl0
   case scanBuf[scanPtr++]
      tokPlus2:
         ' Emit ++ (post-increment)
      tokMinus2:
         ' Emit -- (post-decrement)
      tokSet:
         ' Emit ~~ (set)
      tokClear:
         ' Emit ~ (clear)
      tokRandom:
         ' Emit ? (post-random)
      other:
         scanPtr--

PRI exprLvl0(n) | t
   if n > maxUnary
      abort string("exprLvl0 - Unary ops")
   case scanBuf[scanPtr++]
      tokPlus2:
         exprLvl0(n+1)
         ' Emit ++ (pre-increment)
      tokMinus2:
         exprLvl0(n+1)
         ' Emit -- (pre-decrement)
      tokSet:
         exprLvl0(n+1)
         if checkCons(1)
            ~~stackVal[stackPtr-1]
         else
            t := loadVals(1)
            outputInst($0B,WrR+Imm,always,t,16) ' SHL #16
            outputInst($0E,WrR+Imm,always,t,16) ' SAR #16
      tokClear:
         exprLvl0(n+1)
         if checkCons(1)
            ~stackVal[stackPtr-1]
         else
            t := loadVals(1)
            outputInst($0B,WrR+Imm,always,t,24) ' SHL #24
            outputInst($0E,WrR+Imm,always,t,24) ' SAR #16
      tokRandom:
         exprLvl0(n+1)
         ' Emit ? (pre-random)
      tokAddr:
         exprLvl0(n+1)
         loadAddr(1)
      other:
         scanPtr--
         termLvl1

PRI exprLvl1(n) | t
   if n > maxUnary
      abort string("exprLvl1 - Unary ops")
   case scanBuf[scanPtr++]
      tokPlus:
         exprLvl1(n+1)
      tokMinus:
         exprLvl1(n+1)
         if checkCons(1)
            -stackVal[stackPtr-1]
         else
            t := loadVals(1)
            outputInst($29,WrR,always,t,t) ' NEG
      tokUp2:
         exprLvl1(n+1)
         if checkCons(1)
            ^^stackVal[stackPtr-1]
         else
         ' Emit ^^
      tokBar2:
         exprLvl1(n+1)
         if checkCons(1)
            ||stackVal[stackPtr-1]
         else
         ' Emit ||
      tokEncode:
         exprLvl1(n+1)
         if checkCons(1)
            >|stackVal[stackPtr-1]
         else
         ' Emit >|
      tokDecode:
         exprLvl1(n+1)
         if checkCons(1)
            |<stackVal[stackPtr-1]
         else
         ' Emit |<
      tokBitNot:
         exprLvl1(n+1)
         if checkCons(1)
            !stackVal[stackPtr-1]
         else
            t := loadVals(1)
            outputInst($29,WrR,always,t,t) ' NEG
            outputInst($21,WrR+Imm,always,t,1) ' SUB #1
      other:
         scanPtr--
         exprLvl0(n+1)

PRI exprLvl2 | t
   exprLvl1(0)
   repeat
      case scanBuf[scanPtr++]
         tokSHR:
            exprLvl1(0)
            if checkCons(2)
               stackVal[stackPtr-2] >>= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($0A,WrR,always,t,t-1) ' SHR
            popVal
         tokSHL:
            exprLvl1(0)
            if checkCons(2)
               stackVal[stackPtr-2] <<= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($0B,WrR,always,t,t-1) ' SHL
            popVal
         tokSAR:
            exprLvl1(0)
            if checkCons(2)
               stackVal[stackPtr-2] ~>= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($0E,WrR,always,t,t-1) ' SAR
            popVal
         tokROR:
            exprLvl1(0)
            if checkCons(2)
               stackVal[stackPtr-2] ->= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($08,WrR,always,t,t-1) ' ROR
            popVal
         tokROL:
            exprLvl1(0)
            if checkCons(2)
               stackVal[stackPtr-2] <-= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($09,WrR,always,t,t-1) ' ROL
            popVal
         tokBitRev:
            exprLvl1(0)
            if checkCons(2)
               stackVal[stackPtr-2] ><= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($0F,WrR,always,t,t-1) ' REV
            popVal
         other:
            scanPtr--
            quit
         
PRI exprLvl3 | t
   exprLvl2
   repeat
      case scanBuf[scanPtr++]
         tokBitAnd:
            exprLvl2
            if checkCons(2)
               stackVal[stackPtr-2] &= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($18,WrR,always,t,t-1) ' AND
            popVal
         tokAndNot:
            exprLvl2
            if checkCons(2)
               stackVal[stackPtr-2] &= !stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($19,WrR,always,t,t-1) ' ANDN
            popVal
         other:
            scanPtr--
            quit

PRI exprLvl4 | t
   exprLvl3
   repeat
      case scanBuf[scanPtr++]
         tokBar:
            exprLvl3
            if checkCons(2)
               stackVal[stackPtr-2] |= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($1A,WrR,always,t,t-1) ' OR
            popVal
         tokBitXor:
            exprLvl3
            if checkCons(2)
               stackVal[stackPtr-2] ^= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($1B,WrR,always,t,t-1) ' XOR
            popVal
         other:
            scanPtr--
            quit

PRI exprLvl5
   exprLvl4
   repeat
      case scanBuf[scanPtr++]
         tokStar:
            exprLvl4
            if checkCons(2)
               stackVal[stackPtr-2] *= stackVal[stackPtr-1]
            else
            popVal
            ' Emit *
         tokStar2:
            exprLvl4
            if checkCons(2)
               stackVal[stackPtr-2] **= stackVal[stackPtr-1]
            else
            popVal
            ' Emit **
         tokSlash:
            exprLvl4
            if checkCons(2)
               stackVal[stackPtr-2] /= stackVal[stackPtr-1]
            else
            popVal
            ' Emit /
         tokSlash2:
            exprLvl4
            if checkCons(2)
               stackVal[stackPtr-2] //= stackVal[stackPtr-1]
            else
            popVal
            ' Emit //
         other:
            scanPtr--
            quit

PRI exprLvl6 | t
   exprLvl5
   repeat
      case scanBuf[scanPtr++]
         tokPlus:
            exprLvl5
            if checkCons(2)
               stackVal[stackPtr-2] += stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($20,WrR,always,t,t-1) ' ADD
            popVal
         tokMinus:
            exprLvl5
            if checkCons(2)
               stackVal[stackPtr-2] -= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($21,WrR,always,t,t-1) ' SUB
            popVal
         other:
            scanPtr--
            quit

PRI exprLvl7
   exprLvl6
   repeat
      case scanBuf[scanPtr++]
         tokMaxOp:
            exprLvl6
            if checkCons(2)
               stackVal[stackPtr-2] #>= stackVal[stackPtr-1]
            else
            popVal
            ' Emit #>
         tokMinOp:
            exprLvl6
            if checkCons(2)
               stackVal[stackPtr-2] <#= stackVal[stackPtr-1]
            else
            popVal
            ' Emit <#
         other:
            scanPtr--
            quit

PRI exprLvl8 | t
   exprLvl7
   repeat
      case scanBuf[scanPtr++]
         tokLSS:
            exprLvl7
            if checkCons(2)
               stackVal[stackPtr-2] <= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($21,WrC,always,t,t-1) ' CMP WC
               outputInst($1C,WrR,always,t,lmmOnes) ' MUXC
            popVal
         tokGTR:
            exprLvl7
            if checkCons(2)
               stackVal[stackPtr-2] >= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($21,WrC,always,t-1,t) ' CMP WC
               outputInst($1C,WrR,always,t,lmmOnes) ' MUXC
            popVal
         tokNEQ:
            exprLvl7
            if checkCons(2)
               stackVal[stackPtr-2] <>= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($21,WrZ,always,t,t-1) ' CMP WZ
               outputInst($1F,WrR,always,t,lmmOnes) ' MUXNZ
            popVal
         tokEQL:
            exprLvl7
            if checkCons(2)
               stackVal[stackPtr-2] === stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($21,WrZ,always,t,t-1) ' CMP WZ
               outputInst($1E,WrR,always,t,lmmOnes) ' MUXZ
            popVal
         tokLEQ:
            exprLvl7
            if checkCons(2)
               stackVal[stackPtr-2] =<= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($21,WrC,always,t-1,t) ' CMP WC
               outputInst($1D,WrR,always,t,lmmOnes) ' MUXNC
            popVal
         tokGEQ:
            exprLvl7
            if checkCons(2)
               stackVal[stackPtr-2] =>= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($21,WrC,always,t,t-1) ' CMP WC
               outputInst($1D,WrR,always,t,lmmOnes) ' MUXNC
            popVal
         other:
            scanPtr--
            quit

PRI exprLvl9 | doNot, t
   doNot := false
   repeat while scanBuf[scanPtr++] == tokNOT
      NOT doNot
   scanPtr--
   exprLvl8
   if doNot
      if checkCons(1)
         NOT stackVal[stackPtr-1]
      else
         t := loadVals(1)
         outputInst($18,WrZ,always,t,lmmOnes) ' TEST
         outputInst($1E,WrR,always,t,lmmOnes) ' MUXZ

PRI exprLvl10 | t
   exprLvl9
   repeat
      case scanBuf[scanPtr++]
         tokAND:
            exprLvl10
            if checkCons(2)
               stackVal[stackPtr-2] AND= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($18,WrZ,always,t,lmmOnes)   ' TEST
               outputInst($1F,WrR,always,t,lmmOnes)   ' MUXNZ
               outputInst($18,WrZ,always,t-1,lmmOnes) ' TEST
               outputInst($1F,WrR,always,t-1,lmmOnes) ' MUXNZ
               outputInst($18,WrZ,always,t,t-1)       ' AND
            popVal
         other:
            scanPtr--
            quit

PRI exprLvl11 | t
   exprLvl10
   repeat
      case scanBuf[scanPtr++]
         tokOR:
            exprLvl10
            if checkCons(2)
               stackVal[stackPtr-2] OR= stackVal[stackPtr-1]
            else
               t := loadVals(2)
               outputInst($18,WrZ,always,t,lmmOnes)   ' TEST
               outputInst($1F,WrR,always,t,lmmOnes)   ' MUXNZ
               outputInst($18,WrZ,always,t-1,lmmOnes) ' TEST
               outputInst($1F,WrR,always,t-1,lmmOnes) ' MUXNZ
               outputInst($18,WrZ,always,t,t-1)       ' AND
            popVal
         other:
            scanPtr--
            quit

PRI assignment | t
   ifnot variable
      return false
   if scanBuf[scanPtr++] <> tokEQL
      return false
   exprLvl11
   storeVal(2)
   popVal
   popVal
   return true

PRI statement(n) | t, i
   scanPtr -= n                  ' Back up to beginning of statement
   case t := scanBuf[scanPtr++]
      tokLET:
         ifnot assignment
            abort string("statement - Syntax error")
      tokIF:
         exprLvl11
         if scanBuf[scanPtr++] <> tokTHEN
            abort string("statement - Missing THEN")
         t := loadVals(1)
         outputInst($1A,WrZ+Imm,always,t,0) ' OR #0
         goForward(ifZ,2)
         t := getLabel
         dictChn[t] := longJump(dictChn[t])
      tokGOTO:
         t := getLabel
         dictChn[t] := longJump(dictChn[t])
      tokGOSUB:
         t := getLabel
         dictChn[t] := callRoutine(dictChn[t])
      tokRETURN:
         outputInst($17,0,always,0,lmmRetn) ' JMP lmmRetn
      tokWAITCNT:
         exprLvl11
         t := loadVals(1)
         outputInst($3E,Imm,always,t,0)
      tokEND:
         compiling := false
      tokINCLUDE:
         case scanBuf[scanPtr++]
            tokChar:
               abort string("statement - Filename < 5 chars")
            tokStr..tokStr+15:
               t := (t-tokStr) << 8 + scanBuf[scanPtr++]
               repeat i from 0 to maxIncl-1
                  if t&$FFF == includes[i]&$FFF
                     return
                  if t == 0
                     includes[i] := $1000 | t
                     return
               abort string("statement - Too many include files")
            other:
               abort string("statement - No include file")
      tokEOL, tokEOF:
         scanPtr--
      other:
         scanPtr--
         ifnot assignment
            abort string("statement - Syntax error")

PRI processLine(len) | t, s, z
   scanPtr := 0
   stackPtr := 0
   case t := scanBuf[scanPtr++]
      tokIdent..tokLastId, tokExtId..tokExtId+15:
         if t < tokExtId
            t -= tokIdent
         else
            t -= tokExtId
            t := t << 8 + scanBuf[scanPtr++]
            t += tokLastId - tokIdent + 1
         case scanBuf[scanPtr++]
            tokCON:
               exprLvl11
               z := mustBeVal
               if findSymbol(symAny,t) == symAny
                  t := findSymbol(symCons,t)
                  dictLen[t] := z & $FFFF
                  dictChn[t] := z >> 16
               else
                  t := findSymbol(symCons,t)
                  if t == clkmodePtr and !clkmodeSet
                     dictLen[t] := z & $FFFF
                     dictChn[t] := z >> 16
                     clkmodeSet := true
                  elseif t == clkfreqPtr and !clkfreqSet
                     dictLen[t] := z & $FFFF
                     dictChn[t] := z >> 16
                     clkfreqSet := true
                  else
                     abort string("processLine - Constant already defined")
            tokVAR:
               case scanBuf[scanPtr++]
                  tokBYTE:
                     s := symByte
                  tokWORD:
                     s := symWord
                  tokLONG:
                     s := symLong
                  tokLOCAL:
                     s := symLocl
                  other:
                     abort string("processLine - VAR type ?")
               if scanBuf[scanPtr++] == tokLtPar
                  exprLvl11
                  if scanBuf[scanPtr++] <> tokRtPar
                     abort string("processLine - Syntax error")
                  z := mustBeVal
                  if z < 1 or z > 32767 or s == symLocl
                     abort string("processLine - Invalid dimension")
               else
                  scanPtr--
                  z := 1
               if findSymbol(symAny,t) == symAny
                  t := findSymbol(s,t)
                  dictLen[t] := z
                  if s == symLocl
                     dictChn[t] := localVars++
                     if localVars > lastLocal
                        abort string("processLine - Too many local variables")
               else
                  abort string("processLine - Variable already defined")
            tokColon:
               case findSymbol(symAny,t)
                  symLabel:
                     t := findSymbol(symLabel,t)
                     if dictChn[t] & $8000 == 0
                        abort string("processLine - Label already defined")
                     fixupChain(dictChn[t]&$7FFF,lib.getOutputAddr)
                     dictChn[t] := lib.getOutputAddr | $8000
                  symAny:
                     t := findSymbol(symLabel,t)
                     dictChn[t] := lib.getOutputAddr | $8000
                  other:
                     abort string("processLine - Identifier not label")
            other:
               statement(2)
      other:
         statement(1)
   if scanBuf[scanPtr] <> tokEOL and scanBuf[scanPtr] <> tokEOF
      abort string("processLine - End of line missing")
   if stackPtr > 0
      abort string("processLine - Stack not empty")

PRI Ouroboros | i, addr, chksum, temp, undefLbl
   lib.message(string("Ouroboros "))
   lib.messageDec(version)
   lib.messageOut(".")
   if release < 100
      lib.messageOut("0")
   if release < 10
      lib.messageOut("0")
   lib.messageDec(release)
   if testLevel > 0
      lib.messageOut("a"+testLevel-1)
   lib.message(string(lib#crChar,lib#lfChar))
   wordfill(@includes,0,maxIncl) ' Clear include files
   lib.setIgnoreCase(true)       ' Set scan modes
   lib.setCommentType("S")
   lib.setLineOriented(true)
   dictPtr := 0                  ' Initialize dictionary and
   predefinedNames               '  install predefined names
   localVars := locals           ' Start of local variables
   temp := (@end_prolog - @bgn_prolog) >> 2
   repeat i from 0 to temp-1     ' Copy Spin prolog to start of EEPROM
      lib.outputLong(bgn_prolog[i])
   clkmodeSet := false
   clkfreqSet := false
   compiling := true

   lib.message(string("Begin compilation",lib#crChar,lib#lfChar))
   repeat while compiling        ' Go through the program line by line
      scanPtr := 0
      repeat                     ' Scan all tokens into a buffer
         if (temp := lib.getOperator(@operators)) => 0
            scanBuf[scanPtr++] := temp + tokPlus2
            next
         if lib.getNumber        ' Do we have a number?  Copy to buffer
            scanBuf[scanPtr++] := tokNumb
            repeat i from 0 to 3
               temp := lib.getValue
               scanBuf[scanPtr++] := temp.byte[i]
            next
         if lib.getContent       ' Do we have a string?  Copy to buffer
            if (temp := lib.getLength) < 5 ' Short strings (< 5 chars)
               scanBuf[scanPtr++] := tokChar
               bytemove(@scanBuf+scanPtr,lib.getSymBuf,temp)
               scanPtr += temp
               scanBuf[scanPtr++] := 0
               next
            ifnot (temp := lib.getSymbol(0)) < 4096 ' Long - treat as symbols
               abort string("Ouroboros - too many symbols")
            scanBuf[scanPtr++] := tokStr + (temp >> 8)
            scanBuf[scanPtr++] := temp & $FF
            next
         if lib.getIdent         ' Do we have an identifier?
            if (temp := lib.getKeyword(@keywords,0)) => 0
               scanBuf[scanPtr++] := temp + tokCON
               next
            if (temp := lib.getSymbol(0)) > (tokLastId - tokIdent + 1)
               ifnot (temp -= tokLastId - tokIdent + 1) < 4096
                  abort string("Ouroboros - too many symbols")
               scanBuf[scanPtr++] := (temp >> 8) + tokExtId
               scanBuf[scanPtr++] := temp & $FF
            else
               scanBuf[scanPtr++] := tokIdent + temp
            next
         if lib.getEndline       ' Is the the end of the line?
            scanBuf[scanPtr] := tokEOL
            quit
         if lib.getEndfile       ' Is this the end of the file?
            scanBuf[scanPtr] := tokEOF
            compiling := false
            quit
         abort string("Ouroboros - syntax error")
      processLine(scanPtr)

   lib.message(string("Finishing up program",lib#crChar,lib#lfChar))
   ' Add stop to end of program
   outputInst($03,WrR+Imm,always,locals-1,$01) ' COGID   +stack
   outputInst($03,Imm,always,locals-1,$03)     ' COGSTOP stack-

   ' Go through the dictionary, allocating storage for string constants,
   ' then the program's global (hub) variables, first bytes, then words,
   ' then longs, all aligned appropriately.  Check for undefined labels.
   undefLbl := false
   if dictPtr > 0
      repeat i from 0 to dictPtr-1
         if dictSym[i] >> 12 == symLabel
            lib.outputSymbol(dictSym[i]&$FFF,false)
            if dictChn[i] & $8000 == 0
               lib.message(string("?????"))
               undefLbl := true
            else
               lib.message(string("$"))
               lib.messageHex(dictChn[i]&$7FFF,4)
            lib.message(string(" Label "))
            lib.message(lib.getSymBuf)
            lib.messageNewline
   if dictPtr > 0
      repeat i from 0 to dictPtr-1
         if dictSym[i] >> 12 == symStr
            lib.message(string("$"))
            lib.messageHex(lib.getOutputAddr,4)
            lib.message(string(" String ",lib#quote))
            fixupChain(dictChn[i],lib.getOutputAddr)
            lib.outputSymbol(dictSym[i]&$FFF,true)
            lib.message(lib.getSymBuf)
            lib.message(string(lib#quote,lib#crChar,lib#lfChar))
   addr := lib.getOutputAddr     ' Start allocating after strings
   if dictPtr > 0
      repeat i from 0 to dictPtr-1
         if dictSym[i] >> 12 == symByte
            if dictLen[i] > 0
               fixupChain(dictChn[i],addr)
               lib.outputSymbol(dictSym[i]&$FFF,false)
               lib.message(string("$"))
               lib.messageHex(addr,4)
               lib.message(string(" Byte "))
               lib.message(lib.getSymBuf)
               lib.messageNewline
               addr += dictLen[i]
   addr := (addr + 1) & !1       ' Align to next word boundary
   if dictPtr > 0
      repeat i from 0 to dictPtr-1
         if dictSym[i] >> 12 == symWord
            if dictLen[i] > 0
               fixupChain(dictChn[i],addr)
               lib.outputSymbol(dictSym[i]&$FFF,false)
               lib.message(string("$"))
               lib.messageHex(addr,4)
               lib.message(string(" Word "))
               lib.message(lib.getSymBuf)
               lib.messageNewline
               addr += dictLen[i] * 2
   addr := (addr + 3) & !3       ' Align to next long boundary
   if dictPtr > 0
      repeat i from 0 to dictPtr-1
         if dictSym[i] >> 12 == symLong
            if dictLen[i] > 0
               fixupChain(dictChn[i],addr)
               lib.outputSymbol(dictSym[i]&$FFF,false)
               lib.message(string("$"))
               lib.messageHex(addr,4)
               lib.message(string(" Long "))
               lib.message(lib.getSymBuf)
               lib.messageNewline
               addr += dictLen[i] * 4
   if undefLbl
      abort string("One or more undefined labels found")
   repeat while (lib.getOutputAddr & 3 <> 0)
      lib.outputByte(0)          ' Round up code to multiple of 4

   ' Update vbase with the total length of the program
   lib.writeLong(8, lib.readLong(8) | lib.getOutputAddr)   

   ' Compute the checksum for the program
   chksum := 0
   repeat i from 0 to lib.getOutputAddr - 4 step 4
      temp := lib.readLong(i)
      chksum -= temp.byte[0]+temp.byte[1]+temp.byte[2]+temp.byte[3]         

   ' Store the clock frequency in the program prolog
   lib.writeLong(0, dictChn[clkfreqPtr] << 16 | dictLen[clkfreqPtr])   

   ' Store the clock mode and checksum in the program prolog
   i := lib.readLong(4) & $FFFF0000 ' Get the rest of the long value
   lib.writeLong(4,i | dictLen[clkmodePtr]&$00FF | (chksum<<8)&$FF00)
   lib.message(string("End of compilation",lib#crChar,lib#lfChar))

PRI goForward(flags,offset)
   outputInst($20, WrR+Imm, flags, PC, offset) ' ADD PC,#offset

PRI goBackward(flags,offset)
   outputInst($21, WrR+Imm, flags, PC, offset) ' SUB PC,#offset

PRI loadImmed(dest,value)
   outputInst($17,0,always,dest,lmmLoad) ' JMP dest,lmmLoad
   outputLong(value)         ' Note: dest used by LMM interpreter

PRI callRoutine(address)
   outputInst($17,0,always,0,lmmCall) ' JMP lmmCall
   outputLong(address & $7FFF)
   if address & $8000 <> 0       ' If already defined,
      return address             '  return current address
   else
      return lib.getOutputAddr-4 ' Otherwise, chain addresses

PRI longJump(address)
   outputInst($17,0,always,PC,lmmLoad) ' JMP PC,lmmLoad
   outputLong(address & $7FFF)
   if address & $8000 <> 0       ' If already defined,
      return address             ' return current address
   else
      return lib.getOutputAddr-4 ' Otherwise, chain addresses

PRI fixupChain(chain,address)    ' Fixup the linked list at chain
   repeat while chain <> 0       '  with the value given as address
      result := lib.readLong(chain)
      lib.writeLong(chain,address)
      chain := result

PRI findSymbol(symType,symNum) | t, n
   t := symType << 12 | symNum   ' Lookup the supplied symbol and
   repeat while result < dictPtr '  return its dictionary index
      n := dictSym[result]
      if n == t
         quit                    ' symAny will match any identifier type
      if symType == symAny and n >> 12 <> symStr
         if n & $FFF == t & $FFF ' Symbol #s have to agree
            quit
      result++
   if symType == symAny          ' If specified symType is symAny,
      if result == dictPtr
         return symAny           '  Return symAny if symbol not found
      else
         return n >> 12          '  Return actual symbol type if found
   if result == dictPtr
      if result => maxDict
         abort string("findSymbol - Too many symbols")
      dictSym[dictPtr] := t      ' Create new dictionary entry
      dictLen[dictPtr] := 0
      dictChn[dictPtr++] := 0

PRI installSym(s,t,vl,vh)        ' Install a predefined symbol
   dictSym[dictPtr] := t << 12 | lib.getSymbol(s)
   dictLen[dictPtr] := vl
   dictChn[dictPtr] := vh
   return dictPtr++

PRI predefinedNames
   installSym(string("TRUE"),symCons,$FFFF,$FFFF)
   installSym(string("FALSE"),symCons,$0000,$0000)
   installSym(string("CNT"),symLocl,0,$1F1+maxDict+1)
   installSym(string("INA"),symLocl,0,$1F2+maxDict+1)
   installSym(string("OUTA"),symLocl,0,$1F4+maxDict+1)
   installSym(string("DIRA"),symLocl,0,$1F6+maxDict+1)
   installSym(string("CTRA"),symLocl,0,$1F8+maxDict+1)
   installSym(string("CTRB"),symLocl,0,$1F9+maxDict+1)
   installSym(string("FRQA"),symLocl,0,$1FA+maxDict+1)
   installSym(string("FRQB"),symLocl,0,$1FB+maxDict+1)
   installSym(string("PHSA"),symLocl,0,$1FC+maxDict+1)
   installSym(string("PHSB"),symLocl,0,$1FD+maxDict+1)
   installSym(string("VCFG"),symLocl,0,$1FE+maxDict+1)
   installSym(string("VSCL"),symLocl,0,$1FF+maxDict+1)
   installSym(string("POSX"),symCons,$FFFF,$7FFF)
   installSym(string("NEGX"),symCons,$0000,$8000)
   installSym(string("RCFAST"),symCons,%0_0_0_00_000,0)
   installSym(string("RCSLOW"),symCons,%0_0_0_00_001,0)
   installSym(string("XINPUT"),symCons,%0_0_0_00_010,0)
   installSym(string("XTAL1"),symCons,%0_0_1_01_010,0)
   installSym(string("XTAL2"),symCons,%0_0_1_10_010,0)
   installSym(string("XTAL3"),symCons,%0_0_1_11_010,0)
   installSym(string("PLL1X"),symCons,%0_1_0_00_001,0)
   installSym(string("PLL2X"),symCons,%0_1_0_00_010,0)
   installSym(string("PLL4X"),symCons,%0_1_0_00_011,0)
   installSym(string("PLL8X"),symCons,%0_1_0_00_100,0)
   installSym(string("PLL16X"),symCons,%0_1_0_00_101,0)
   installSym(string("CLKMODE"),symByte,0,4+maxDict+1)
   installSym(string("CLKFREQ"),symLong,0,0+maxDict+1)
   clkmodePtr := installSym(string("_CLKMODE"),symCons,%0_0_0_00_000,0)
   clkfreqPtr := installSym(string("_CLKFREQ"),symCons,initClk&$FFFF,initClk>>16)

PRI outputLong(val)
   lib.message(string("Long "))
   lib.messageHex(lib.getOutputAddr,4)
   lib.message(string(" -> "))
   lib.messageHex(val,8)
   lib.message(string(" / "))
   lib.messageDec(val)
   lib.messageNewline
   lib.outputLong(val)

PRI outputInst(opCode,effect,flags,dest,src)
   lib.message(string("Inst "))
   lib.messageHex(lib.getOutputAddr,4)
   lib.message(string(" -> "))
   lib.messageHex(opCode,2)
   lib.message(string(" / "))
   lib.messageHex(effect,1)
   lib.message(string(" / "))
   lib.messageHex(flags,1)
   lib.message(string(" / "))
   lib.messageHex(dest,3)
   lib.message(string(" / "))
   lib.messageHex(src,3)
   lib.messageNewline
   lib.outputInst(opCode,effect,flags,dest,src)

DAT
operators byte
 byte "++",0, "--",0, "^^",0, "||",0, "**",0, "//",0, "#>",0, "<#",0
 byte "~>",0, "~~",0, "<<",0, ">>",0, "->",0, "<-",0, "|<",0, ">|",0
 byte "><",0, ">=",0, "<=",0, "<>",0, "..",0, "&/",0, "??",0
 byte "+",0, "-",0, "^",0, "|",0, "*",0, "/",0, "~",0, "?",0, "@",0
 byte "!",0, ">",0, "<",0, "&",0, ":",0, "(",0, ")",0, "[",0, "]",0
 byte ",",0, ";",0, "#",0, "=",0, ".",0, "\",0
 byte 0

keywords byte
 byte "CON",0, "VAR",0, "BYTE",0, "WORD",0, "LONG",0, "OR",0, "AND",0
 byte "NOT",0, "BRANCH",0, "IF",0, "THEN",0, "ELSE",0, "ENDIF",0, "GOTO",0
 byte "GOSUB",0, "RETURN",0, "FOR",0, "TO",0, "STEP",0, "NEXT",0, "READ",0
 byte "WRITE",0, "LET",0, "LOOKUP",0, "LOOKDOWN",0, "INPUT",0, "OUTPUT",0
 byte "REVERSE",0, "LOW",0, "HIGH",0, "TOGGLE",0, "PAUSE",0, "WAITCNT",0
 byte "WAITPNE",0, "WAITPEQ",0, "WAITVID",0, "END",0, "DEBUG",0, "CNT",0
 byte "INA",0, "OUTA",0, "DIRA",0, "CTRA",0, "CTRB",0, "FRQA",0, "FRQB",0
 byte "PHSA",0, "PHSB",0, "VCFG",0, "VSCL",0, "REBOOT",0, "LOCKSET",0
 byte "LOCKCLR",0, "LOCKNEW",0, "LOCKRET",0, "COGNEW",0, "COGINIT",0
 byte "COGID",0, "COGSTOP",0, "CLKSET",0, "LOCAL",0, "INCLUDE",0
 byte 0

bgn_prolog long
' The following is the Spin control block starting at location 0
' It causes the LMM kernel following it to be loaded into cog 0
' and started beginning with the first LMM instruction at location
' $00C4 which immediately follows the kernel code.

' The checksum in byte $0005 and the length of the program in the
' word at $0004 need to be updated before writing this to a file.
' The length of the program should be a multiple of 4.  The checksum
' is simply the two's complement of the sum of the bytes of the
' program file so the sum of the whole file will come out to zero.

 long $00000000  ' value is clkfreq = 12_000_000
 long $00100000  ' 2nd byte is checksum, _clkmode = rcfast
 long $00180000  ' 1st word is vbase = total size as multiple of 4
 long $00100018 
 long $FFF9FFFF  ' below local stack
 long $FFF9FFFF
 long $39043735  ' push #0; push #$20; push #$00C4
 long $002C00C4  ' coginit

                 '   f_temp    = $031               ' Temporary variables
                 '   f_temp2   = $032
                 '   cacheArea = $040               ' Start of cache area
                 '   endCache  = $080               ' Next location past cache
                 '   endStack  = $1F0               ' Next location past stack
                 '   PC        = 0                  ' LMM program counter
                 '   SP        = 1                  ' LMM stack pointer
                 '   lmmNext   = 2                  ' LMM next instruction
                 '   lmmCall   = 3                  ' LMM subroutine call
                 '   lmmRetn   = 4                  ' LMM subroutine return
                 '   lmmCache  = 5                  ' LMM load/execute cache
                 '   lmmLoad   = 6                  ' LMM load next long
                 '   lmmExp    = 7                  ' LMM expansion routine

 long $5c7c0008  '00           jmp     #f_initial   ' Program counter
 long $00000000  '01           long    0            ' Hub RAM stack pointer
 long $0000001c  '02           long    f_next       ' Next instruction routine
 long $00000016  '03           long    f_call       ' Subroutine call routine
 long $00000013  '04           long    f_retn       ' Subroutine return routine
 long $0000000b  '05           long    f_cache      ' Load/execute cache routine
 long $00000029  '06           long    f_load       ' Load next long routine
 long $0000001c  '07           long    f_next       ' Reserved for expansion
                    
 long $a0bc01f0  '08 f_initial mov     PC,PAR       ' Set initial PC value
 long $50fc27f0  '09           movs    f_retn,#endStack
 long $5c7c001c  '0a           jmp     #f_next      ' Set top of stack
                    
 long $08bc6200  '0b f_cache   rdlong  f_temp,PC    ' Get number of longs
 long $80fc0004  '0c :incrPC   add     PC,#4
 long $08bc8000  '0d :fcache   rdlong  cacheArea,PC ' Copy to the cache
 long $80bc1a12  '0e           add     :fcache,:incr2
 long $e4fc620c  '0f           djnz    f_temp,#:incrPC
 long $54fc1a40  '10           movd    :fcache,#cacheArea
 long $5c7c0040  '11           jmp     #cacheArea   ' Execute the cached code
                    
 long $00000200  '12 :incr2    long    $00000200    ' Increment destination
                    
 long $a0bc0000  '13 f_retn    mov     PC,0         ' Pop the cog call stack
 long $80fc2601  '14           add     f_retn,#1
 long $5c7c001c  '15           jmp     #f_next
                    
 long $08bc6200  '16 f_call    rdlong  f_temp,PC    ' Push return address
 long $84fc2601  '17           sub     f_retn,#1    '  onto the call stack
 long $54bc3413  '18           movd    :storePc,f_retn
 long $80fc0004  '19           add     PC,#4
 long $a0bc0000  '1a :storePc  mov     0,PC
 long $a0bc0031  '1b           mov     PC,f_temp    ' Assumes f_next follows
                    
 long $08bc3c00  '1c f_next    rdlong  :inst1,PC    ' Next instruction
 long $80fc0004  '1d           add     PC,#4
 long $00000000  '1e :inst1    nop
 long $08bc4200  '1f           rdlong  :inst2,PC
 long $80fc0004  '20           add     PC,#4
 long $00000000  '21 :inst2    nop
 long $08bc4800  '22           rdlong  :inst3,PC
 long $80fc0004  '23           add     PC,#4
 long $00000000  '24 :inst3    nop
 long $08bc4e00  '25           rdlong  :inst4,PC
 long $80fc0004  '26           add     PC,#4
 long $00000000  '27 :inst4    nop
 long $5c7c001c  '28           jmp     #f_next
                    
 long $08bc6200  '29 f_load    rdlong  f_temp,PC    ' Getthedata from the
 long $84fc0004  '2a           sub     PC,#4        '  next long and store
 long $08bc6400  '2b           rdlong  f_temp2,PC   '   according to the
 long $28fc6409  '2c           shr     f_temp2,#9   '    destination field
 long $54bc5e32  '2d           movd    :loadIt,f_temp2
 long $80fc0008  '2e           add     PC,#8
 long $a0bc0031  '2f :loadIt   mov     0,f_temp
 long $5c7c001c  '30           jmp     #f_next

                org $033
' 
' Multiply x[15..0] by y[15..0] (y[31..16] must be 0) 
' on exit, product in y[31..0] 
' 
multiply        shl    $7F,#16           'get multiplicand into x[31..16] 
                mov    $7D,#16           'ready for 16 multiplier bits 
                shr    $7E,#1    wc      'get initial multiplier bit into c 
 
:loop   if_c    add    $7E,$7F   wc      'if c set, add multiplicand into product 
                rcr    $7E,#1    wc      'get next multiplier bit into c, shift product 
                djnz   $7D,#:loop        'loop until done 
 
multiply_ret    ret                      'return with product in y[31..0]

' 
' Divide x[31..0] by y[15..0] (y[16] must be 0) 
' on exit, quotient is in x[15..0] and remainder is in x[31..16] 
' 
divide          shl    $7E,#15           'get divisor into y[30..15] 
                mov    $7D,#16           'ready for 16 quotient bits 
 
:loop           cmpsub $7F,$7E   wc      'if y =< x then subtract it, quotient bit into c 
                rcl    $7F,#1            'rotate c into quotient, shift dividend 
                djnz   $7D,#:loop        'loop until done 
 
divide_ret      ret                      'quotient in x[15..0], remainder in x[31..16]
end_prolog long