'' Ouroboros system by Michael Green (c) 2007, 2008.
'' Portions (c) Parallax, Inc. as noted in the comments.

'' This is a collection of programs that provides a simple
'' text editor and compiler for a Basic compiler for the
'' Propeller processor developed by Parallax, Inc.

'' This requires a Propeller Protoboard with the Accessory
'' Kit along with a PS/2 keyboard and VGA display.  An SD
'' card socket must be wired to I/O pins 8-11 as described in
'' Definitions.spin.  The VGA monitor should support a
'' 1024 x 768 pixel resolution with a refresh rate of 60Hz.

OBJ
   def  : "Definitions"          ' Common definitions
   data : "MassStorage"          ' Input file interface
   work : "MassStorage"          ' Work file interface
   key  : "Keyboard"             ' Keyboard I/O routines
   dsp  : "Display"              ' Display I/O routines
   vga  : "VgaDriver"            ' VGA driver
   mass : "I2cSpiInit"           ' I2C / SPI driver

CON
   _xinfreq  = def#xtalClock
   _clkmode  = def#clockMode
   _free     = def#endMemory-def#endFree + vga#totalTiles*2 + vga#totalColors*4

CON
   version   = 1                 ' Current version number
   release   = 001               ' Current release number
   testLevel = 1                ' Current test level

   maxFileNm = 12                ' Filename length
   maxLine   = 80                ' Maximum line length for input

DAT
addToBgn byte def#MoveXY,def#Alt00,def#Alt00,0
addToEnd byte def#MoveXY,def#Alt00,23,0

VAR
   long fileBackup               ' One character look behind
   long direction
   byte buffer[maxLine+1]        ' Line buffer

PUB Ouroboros | banner, t        ' Initialize Ouroboros
   banner := not def.resumed     '  Check if already initialized
   def.start                     '  High memory work area
   vga.start(def#vgaPins)        '  VGA display driver (16x32 tiles)
   dsp.start                     '  Display interface
   key.start(def#keyPins)        '  PS/2 keyboard driver
   if banner
      dsp.str(string("Ouroboros "))
      dsp.dec(version)
      dsp.out(".")
      if release < 100
         dsp.out("0")
      if release < 10
         dsp.out("0")
      dsp.dec(release)
      if testLevel > 0
         dsp.out("a"+testLevel-1)
      dsp.str(string(def#Cr,def#Lf))
   ifnot mass.start              ' Start I2C / SPI driver
      dsp.str(string("Can't initialize I2C / SPI",def#Cr,def#Lf))
      abort
   waitcnt(2*clkfreq + cnt)
   ifnot key.present             ' Make sure keyboard present
      dsp.str(string("Keyboard missing",def#Cr,def#Lf))
      abort
   dsp.dec(long[def#memPtr]-@t)
   dsp.str(string(" Free, $"))
   dsp.hex(data.debugAddr>>16,4)
   dsp.str(string(" Globals, $"))
   dsp.hex(data.debugAddr&$FFFF,4)
   dsp.str(string(" Data, $"))
   dsp.hex(work.debugAddr&$FFFF,4)
   dsp.str(string(" Work",def#Cr,def#Lf))
   if t := \data.mountSDVol      ' Attempt to mount SD card
      dsp.dec(t)
      dsp.str(string(" SD card missing",def#Cr,def#Lf))
      dumpMemory(-1,data.debugAddr>>16,80)
      dumpMemory(-1,data.debugAddr&$FFFF,80)
      abort
   if t := \work.openFile(string("workfile.txt"),"w")
      dsp.dec(t)
      dsp.str(string(" Can't open WORKFILE.TXT",def#Cr,def#Lf))
      dumpMemory(-1,work.debugAddr>>16,80)
      dumpMemory(-1,work.debugAddr&$FFFF,80)
      abort
   repeat                        ' Ask for file name
      askFilename(string("Filename? "))
      if buffer[0] == 0
         data.startSDFiles       ' If none given, list directory
         dsp.str(string("Directory listing",def#Cr,def#Lf))
         repeat until data.nextSDFiles(@buffer) == -1
            dsp.str(@buffer)
            dsp.str(string(def#Cr,def#Lf))
         dsp.str(string("End of directory",def#Cr,def#Lf))
      else
         if data.openFile(@buffer,"r")
            dsp.str(string("Can't open file",def#Cr,def#Lf))
         else
            quit
   direction := true
   repeat
      fileBackup := -1
      result := fileEditor
      not direction
   until result
   reboot

PRI askFilename(s) | i           ' Get file name into inTxt
   dsp.str(s)
   i := 0
   repeat
      case buffer[i] := key.getKey
         def#Bsp, $C8:          ' Backspace
            if i
               dsp.str(string(def#Bsp," ",def#Bsp))
               i--
         def#Cr:                ' Carriage return
            buffer[i]~          ' Zero byte terminator
            dsp.str(string(def#Cr,def#Lf))
            quit
         other:                  ' Other characters,
            if i < maxFileNm     '  save up to maxFilenm
               dsp.out(buffer[i++])

PRI fileEditor | t, theEnd
   theEnd := false
   dsp.str(string(def#Alt00,"Ouroboros Editor",def#Cr))
   data.seekFile(0)
   work.seekFile(0)
   repeat t from 1 to 23
      dsp.out(def#Lf)
      ifnot theEnd
         ifnot theEnd := getFileLine(@buffer)
            dsp.str(@buffer)
      dsp.out(def#Cr)
   t := 1
   repeat
      t := dsp.windowEditor(0,1,64,23,t)
      case t & $FF
         $C6:                    ' Page up
            dsp.getLine(@buffer,0,1,64)
            putFileLine(@buffer)
            dsp.scrollUp(0,1,64,23)
            dsp.str(addToEnd)
            ifnot theEnd
               ifnot theEnd := getFileLine(@buffer)
                  dsp.str(@buffer)
            dsp.out(def#Cr)
         $C7:                    ' Page down
'           dsp.getLine(@buffer,0,23,64)
'           dsp.scrollDown(0,1,64,23)
'           dsp.str(addToBgn)
         $CB:                    ' Escape
            result := false
            quit
         $CC:                    ' App
            result := true
            quit
   repeat t from 1 to 23
      dsp.getLine(@buffer,0,t,64)
      putFileLine(@buffer)
   ifnot theEnd
      repeat until theEnd := getFileLine(@buffer)
         putFileLine(@buffer)

PRI getFileLine(p) | onlyEOF, t  ' Read a line of text
   onlyEOF := true
   repeat
      if fileBackup <> -1
         result := fileBackup~~
      else
         if direction
            result := data.getByte
         else
            result := work.getByte
      case result
         -1, 0:                  ' End of file
            byte[p] := 0
            fileBackup~
            return onlyEOF
         def#Cr:
            onlyEOF := false
            if direction
               fileBackup := data.getByte
            else
               fileBackup := work.getByte
            case fileBackup
               -1, 0:
                  fileBackup~
               def#Lf:
                  fileBackup~~
               other:
                  next
            byte[p] := 0
            return false
         def#Lf:
            byte[p] := 0
            return false
         other:
            byte[p++] := result
            onlyEOF := false

PRI putFileLine(p)
   if direction
      work.str(p)
      work.out(def#Cr)
      work.out(def#Lf)
   else
      data.str(p)
      data.out(def#Cr)
      data.out(def#Lf)

PUB dumpMemory(pin,addr,size) | i, c, p, first, bufa, bufb, bufc
'' This routine dumps a portion of the RAM/ROM to the display (pin == -1).
'' If pin is not -1, it is an EEPROM address in the form required by the
'' I2C routines in i2cSpiInit.  The specified address is or'd with this.
'' The format is 8 bytes wide with hexadecimal and ASCII.
   addr &= $7FFFF
   first := true
   p := addr & $7FFF8
   repeat while p < (addr + size)
      if first
         dsp.hex(addr,5)
         first := false
      else
         dsp.hex(p,5)
      dsp.out(":")
      repeat i from 0 to 7
         byte[@bufa][i] := " "
         if p => addr and p < (addr + size)
            if pin <> -1
               c := 0
               if data.readEEPROM(pin|p,@c,1)
                  abort string("EEPROM read")
            else
               c := byte[p]
            dsp.hex(c,2)
            if c => " " and c =< "~"
               byte[@bufa][i] := c
         else
            dsp.out(" ")
            dsp.out(" ")
         dsp.out(" ")
         p++
      bufc := 0
      dsp.out("|")
      dsp.str(@bufa)
      dsp.out("|")
      dsp.out(def#Cr)
      dsp.out(def#Lf)
            